@@ -230,6 +230,7 @@ backend_build_dn(const char *attribute, const char *value,
 static Slapi_Entry *
 backend_make_user_entry_from_nsswitch_passwd(struct passwd *pwd,
 					     char *container_sdn,
+					     char *user_name,
 					     struct backend_search_cbdata *cbdata)
 {
 	Slapi_Entry *entry;
@@ -272,6 +273,18 @@ backend_make_user_entry_from_nsswitch_passwd(struct passwd *pwd,
 			       "objectClass", "posixAccount");
 	slapi_entry_add_string(entry,
 			       "uid", name);
+	if (user_name != NULL) {
+		/* For non-NULL original user name check if it was
+		 * an alias/UPN. If so, add it to the entry.
+		 * Yes, LDAP schema allows multiple values of 'uid'
+		 * attribute.
+		 */
+		if (slapi_utf8casecmp((unsigned char*) user_name,
+				      (unsigned char*) name) != 0) {
+			slapi_entry_add_string(entry, "uid", user_name);
+		}
+	}
+
 	slapi_entry_attr_set_uint(entry,
 				 "uidNumber", pwd->pw_uid);
 	slapi_entry_attr_set_uint(entry,
@@ -510,6 +523,7 @@ repeat:
 	}
 
 	entry = backend_make_user_entry_from_nsswitch_passwd(&pwd, container_sdn,
+							     is_uid ? NULL : user_name,
 							     cbdata);
 	entries = malloc(sizeof(entries[0]) * 2);
 	if (entries != NULL) {
