@@ -54,6 +54,9 @@
 #include "map.h"
 #include "back-sch.h"
 
+static int backend_passwdmod_extop(Slapi_PBlock *pb);
+backend_extop_handlers_t extop_handlers[] = {{EXTOP_PASSWD_OID, (IFP) backend_passwdmod_extop},
+					    {NULL, NULL}};
 static void
 backend_entries_to_return_push(struct backend_search_cbdata *cbdata, Slapi_Entry *e);
 
@@ -95,6 +98,7 @@ backend_set_config_free_config_contents(void *data)
 		slapi_sdn_free(&set_data->container_sdn);
 		free(set_data->rdn_format);
 		backend_shr_free_strlist(set_data->attribute_format);
+		slapi_ch_free_string(&set_data->associated_domain);
 	}
 }
 void
@@ -146,6 +150,7 @@ backend_copy_set_config(const struct backend_set_data *data)
 	ret->check_access = data->check_access;
 	ret->check_nsswitch = data->check_nsswitch;
 	ret->nsswitch_min_id = data->nsswitch_min_id;
+	ret->associated_domain = data->associated_domain ? slapi_ch_strdup(data->associated_domain) : NULL;
 
 	if ((ret->common.group == NULL) ||
 	    (ret->common.set == NULL) ||
@@ -263,6 +268,37 @@ backend_set_config_read_config(struct plugin_state *state, Slapi_Entry *e,
 		free(nsswitch_min_id);
 	}
 
+	ret.associated_domain = NULL;
+	if (ret.common.restrict_subtrees != NULL) {
+		Slapi_PBlock *pb = NULL;
+		int result = 0;
+		Slapi_Entry **entries = NULL;
+		int i,j;
+		for (i=0; ret.common.restrict_subtrees[i] != NULL; i++) {
+			pb = wrap_pblock_new(NULL);
+			if (pb != NULL) {
+				slapi_search_internal_set_pb_ext(pb, (Slapi_DN*) ret.common.restrict_subtrees[i], LDAP_SCOPE_BASE,
+								 "(&(objectclass=domainRelatedObject)(associatedDomain=*))",
+								 NULL, 0, NULL, NULL, state->plugin_identity, 0);
+				result = slapi_search_internal_pb(pb);
+				slapi_pblock_get(pb, SLAPI_PLUGIN_INTOP_RESULT, &result);
+				if (result == 0) {
+					slapi_pblock_get(pb, SLAPI_PLUGIN_INTOP_SEARCH_ENTRIES, &entries);
+					for (j=0; entries[j] != NULL; j++) {
+						ret.associated_domain = slapi_entry_attr_get_charptr(entries[j], "associatedDomain");
+						if (ret.associated_domain != NULL)
+							break;
+					}
+				}
+				slapi_free_search_results_internal(pb);
+			}
+			slapi_pblock_destroy(pb);
+			pb = NULL;
+			if (ret.associated_domain != NULL)
+				break;
+		}
+	}
+
 	*pret = backend_copy_set_config(&ret);
 	if (*pret == NULL) {
 		if (strlen(container) > 0) {
@@ -434,6 +470,7 @@ backend_set_process_external_members(Slapi_PBlock *pb,
 	struct backend_staged_search staged = {0, };
 	struct backend_search_cbdata cbdata = {0, };
 	char *plugin_id = state->plugin_desc->spd_id;
+	char *gname = NULL;
 
 	is_attr_exists = slapi_entry_attr_find(e, IPA_ATTR_EXTERNAL_MEMBER, &attr) == 0;
 
@@ -445,6 +482,11 @@ backend_set_process_external_members(Slapi_PBlock *pb,
 	 * and update entry's memberUid attribute */
 
 	staged.name = slapi_entry_attr_get_charptr(e, "cn");
+	if (data->associated_domain != NULL) {
+		gname = slapi_ch_smprintf("%s@%s", staged.name, data->associated_domain);
+		slapi_ch_free_string(&staged.name);
+		staged.name = gname;
+	}
 	staged.type = SCH_NSSWITCH_GROUP;
 	staged.search_members = FALSE;
 	staged.is_id = FALSE;
@@ -1610,6 +1652,8 @@ backend_search_cb(Slapi_PBlock *pb)
 	struct backend_search_cbdata cbdata;
 	struct backend_staged_search *staged, *next;
 	int i, isroot, ret;
+	char *original_target = NULL;
+	char *target = NULL;
 
 	if (wrap_get_call_level() > 0) {
 		return 0;
@@ -1634,7 +1678,7 @@ backend_search_cb(Slapi_PBlock *pb)
 		return 0;
 	}
 
-	slapi_pblock_get(pb, SLAPI_SEARCH_TARGET, &cbdata.target);
+	slapi_pblock_get(pb, SLAPI_SEARCH_TARGET, &original_target);
 	slapi_pblock_get(pb, SLAPI_SEARCH_SCOPE, &cbdata.scope);
 	slapi_pblock_get(pb, SLAPI_SEARCH_SIZELIMIT, &cbdata.sizelimit);
 	slapi_pblock_get(pb, SLAPI_SEARCH_TIMELIMIT, &cbdata.timelimit);
@@ -1655,15 +1699,15 @@ backend_search_cb(Slapi_PBlock *pb)
 	/* Okay, we can search. */
 	slapi_log_error(SLAPI_LOG_PLUGIN, cbdata.state->plugin_desc->spd_id,
 			"searching from \"%s\" for \"%s\" with scope %d%s\n",
-			cbdata.target, cbdata.strfilter, cbdata.scope,
+			original_target, cbdata.strfilter, cbdata.scope,
 			backend_sch_scope_as_string(cbdata.scope));
-	cbdata.target_dn = slapi_sdn_new_dn_byval(cbdata.target);
+	cbdata.target_dn = slapi_sdn_new_dn_byval(original_target);
 	/* Check if there's a backend handling this search. */
 	if (!slapi_be_exist(cbdata.target_dn)) {
 		slapi_log_error(SLAPI_LOG_PLUGIN,
 				cbdata.state->plugin_desc->spd_id,
 				"slapi_be_exists(\"%s\") = 0, "
-				"ignoring search\n", cbdata.target);
+				"ignoring search\n", original_target);
 		slapi_sdn_free(&cbdata.target_dn);
 		return 0;
 	}
@@ -1674,22 +1718,23 @@ backend_search_cb(Slapi_PBlock *pb)
 	 * detect the ID view use. Unless the ID view is within the set we control, don't consider the override */
 	map_data_foreach_domain(cbdata.state, backend_search_find_set_dn_cb, &cbdata);
 	if (cbdata.answer == FALSE) {
-		idview_replace_target_dn(&cbdata.target, &cbdata.idview);
+		target = slapi_ch_strdup(original_target);
+		idview_replace_target_dn(&target, &cbdata.idview);
 		if (cbdata.idview != NULL) {
 			slapi_sdn_free(&cbdata.target_dn);
 			/* Perform another check, now for rewritten DN */
-			cbdata.target_dn = slapi_sdn_new_dn_byval(cbdata.target);
+			cbdata.target_dn = slapi_sdn_new_dn_byval(target);
 			map_data_foreach_domain(cbdata.state, backend_search_find_set_dn_cb, &cbdata);
 			/* Rewritten DN might still be outside of our trees */
 			if (cbdata.answer == TRUE) {
 				slapi_log_error(SLAPI_LOG_PLUGIN, cbdata.state->plugin_desc->spd_id,
 						"Use of ID view '%s' is detected, searching from \"%s\" "
 						"for \"%s\" with scope %d%s. Filter may get overridden later.\n",
-						cbdata.idview, cbdata.target, cbdata.strfilter, cbdata.scope,
+						cbdata.idview, target, cbdata.strfilter, cbdata.scope,
 						backend_sch_scope_as_string(cbdata.scope));
 			} else {
 				slapi_sdn_free(&cbdata.target_dn);
-				slapi_ch_free_string(&cbdata.target);
+				slapi_ch_free_string(&target);
 				slapi_ch_free_string(&cbdata.idview);
 				slapi_log_error(SLAPI_LOG_PLUGIN,
 						cbdata.state->plugin_desc->spd_id,
@@ -1697,6 +1742,8 @@ backend_search_cb(Slapi_PBlock *pb)
 						"ignoring search\n");
 				return 0;
 			}
+		} else {
+			slapi_ch_free_string(&target);
 		}
 	}
 	cbdata.answer = FALSE;
@@ -1848,7 +1895,7 @@ backend_search_cb(Slapi_PBlock *pb)
 	}
 	slapi_sdn_free(&cbdata.target_dn);
 	if (cbdata.idview != NULL) {
-		slapi_ch_free_string(&cbdata.target);
+		slapi_ch_free_string(&target);
 	}
 	slapi_ch_free_string(&cbdata.idview);
 #ifdef USE_IPA_IDVIEWS
@@ -1862,7 +1909,6 @@ backend_search_cb(Slapi_PBlock *pb)
 /* Locate the entry for a given DN. */
 struct backend_locate_cbdata {
 	struct plugin_state *state;
-	char *target;
 	Slapi_DN *target_dn;
 
 	struct backend_entry_data *entry_data;
@@ -1911,6 +1957,7 @@ static void
 backend_locate(Slapi_PBlock *pb, struct backend_entry_data **data, const char **group, const char**set)
 {
 	struct backend_locate_cbdata cbdata;
+	char *original_target = NULL;
 
 	slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &cbdata.state);
 	if (cbdata.state->plugin_base == NULL) {
@@ -1918,9 +1965,9 @@ backend_locate(Slapi_PBlock *pb, struct backend_entry_data **data, const char **
 		*data = NULL;
 		return;
 	}
-	slapi_pblock_get(pb, SLAPI_TARGET_DN, &cbdata.target);
+	slapi_pblock_get(pb, SLAPI_TARGET_DN, &original_target);
 
-	cbdata.target_dn = slapi_sdn_new_dn_byval(cbdata.target);
+	cbdata.target_dn = slapi_sdn_new_dn_byval(original_target);
 	cbdata.entry_data = NULL;
 	cbdata.entry_group = NULL;
 	cbdata.entry_set = NULL;
@@ -1930,12 +1977,9 @@ backend_locate(Slapi_PBlock *pb, struct backend_entry_data **data, const char **
 	 * rebuild the target's RDN to use original attribute's value */
 	if (cbdata.entry_data == NULL) {
 		char *idview = NULL;
-		char *target, *original_target;
-		target = original_target = slapi_ch_strdup(cbdata.target);
+		char *target = NULL;
+		target = slapi_ch_strdup(original_target);
 		idview_replace_target_dn(&target, &idview);
-		if (target != original_target) {
-			slapi_ch_free_string(&original_target);
-		}
 		if (idview != NULL) {
 			char *rdnstr;
 			char *val;
@@ -1950,7 +1994,6 @@ backend_locate(Slapi_PBlock *pb, struct backend_entry_data **data, const char **
 					bval.bv_val = slapi_ch_strdup(val);
 					memset(&scbdata, 0, sizeof(scbdata));
 					scbdata.idview = idview;
-					scbdata.target = target;
 					scbdata.pb = pb;
 					scbdata.state = cbdata.state;
 					scbdata.target_dn = slapi_sdn_new_dn_byval(target);
@@ -1983,7 +2026,6 @@ backend_locate(Slapi_PBlock *pb, struct backend_entry_data **data, const char **
  * insufficient-access error. */
 struct backend_group_check_scope_cbdata {
 	struct plugin_state *state;
-	const char *target;
 	Slapi_DN *target_dn;
 	bool_t ours;
 };
@@ -2008,14 +2050,15 @@ static bool_t
 backend_check_scope_pb(Slapi_PBlock *pb)
 {
 	struct backend_group_check_scope_cbdata cbdata;
+	char *original_target = NULL;
 
 	slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &cbdata.state);
 	if (cbdata.state->plugin_base == NULL) {
 		/* The plugin was not actually started. */
 		return FALSE;
 	}
-	slapi_pblock_get(pb, SLAPI_TARGET_DN, &cbdata.target);
-	cbdata.target_dn = slapi_sdn_new_dn_byval(cbdata.target);
+	slapi_pblock_get(pb, SLAPI_TARGET_DN, &original_target);
+	cbdata.target_dn = slapi_sdn_new_dn_byval(original_target);
 	cbdata.ours = FALSE;
 	map_data_foreach_domain(cbdata.state, backend_group_check_scope_cb,
 				&cbdata);
@@ -2223,6 +2266,191 @@ done_with_lock:
 	return ret;
 }
 
+/* This callback handles EXTOP_PASSWD_OID "1.3.6.1.4.1.4203.1.11.1"
+ * If the extop defines a USERID, it sets SLAPI_TARGET_SDN to
+ * the reverse mapping of the USERID.
+ *
+ * If it is not possible to retrieve USERID in the ber
+ * then value of SLAPI_TARGET_SDN is unchanged.
+ *
+ * Else the value of SLAPI_TARGET_SDN is freed and replaced
+ * either by the USERID or the reverse mapping of USERID (if it exists)
+ */
+static int
+backend_passwdmod_extop(Slapi_PBlock *pb)
+{
+	struct backend_entry_data *data;
+	struct plugin_state *state;
+	Slapi_DN *sdn = NULL;
+	char *extopdn;
+	char *ndn;
+	const char *entry_group = NULL;
+	const char *entry_set = NULL;
+	struct berval	*extop_value = NULL;
+	BerElement	*ber = NULL;
+	ber_tag_t	tag = 0;
+        ber_len_t	len = (ber_len_t) -1;
+
+	if (wrap_get_call_level() > 0) {
+		return 0;
+	}
+
+	slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &state);
+	if (state->ready_to_serve == 0) {
+		/* No data to serve yet */
+		goto free_and_return;
+	}
+	/* Retrieve the original DN from the ber request */
+	/* Get the ber value of the extended operation */
+	slapi_pblock_get(pb, SLAPI_EXT_OP_REQ_VALUE, &extop_value);
+	if (!BV_HAS_DATA(extop_value)) {
+		goto free_and_return;
+	}
+
+        if ((ber = ber_init(extop_value)) == NULL) {
+		goto free_and_return;
+	}
+
+	/* Format of request to parse
+	 *
+	 * PasswdModifyRequestValue ::= SEQUENCE {
+	 * userIdentity    [0]  OCTET STRING OPTIONAL
+	 * oldPasswd       [1]  OCTET STRING OPTIONAL
+	 * newPasswd       [2]  OCTET STRING OPTIONAL }
+	 *
+	 * The request value field is optional. If it is
+	 * provided, at least one field must be filled in.
+	 */
+
+	/* ber parse code */
+	if ( ber_scanf( ber, "{") == LBER_ERROR ) {
+		/* The request field wasn't provided.  We'll
+		 * now try to determine the userid and verify
+		 * knowledge of the old password via other
+		 * means.
+		 */
+		goto free_and_return;
+	} else {
+		tag = ber_peek_tag( ber, &len);
+	}
+
+	/* identify userID field by tags */
+	if (tag == LDAP_EXTOP_PASSMOD_TAG_USERID ) {
+
+		if ( ber_scanf( ber, "a", &extopdn) == LBER_ERROR ) {
+			slapi_ch_free_string(&extopdn);
+			goto free_and_return;
+		}
+
+                slapi_log_error(SLAPI_LOG_PLUGIN, "backend_passwdmod_extop",
+			"extopdn = %s\n", extopdn ? extopdn : "<unknown>" );
+
+		/* Free the current target_DN */
+		slapi_pblock_get(pb, SLAPI_TARGET_SDN, &sdn);
+		if (sdn) {
+			const char *olddn;
+			olddn = slapi_sdn_get_ndn(sdn);
+			slapi_log_error(SLAPI_LOG_PLUGIN, "backend_passwdmod_extop",
+							  "olddn = %s (unknown expected)\n", olddn ? olddn : "<unknown>" );
+			slapi_sdn_free(&sdn);
+		}
+
+		/* replace it with the one in the extop req*/
+		sdn = slapi_sdn_new_dn_byref(extopdn);
+		slapi_pblock_set(pb, SLAPI_TARGET_SDN, sdn);
+	} else {
+		/* we can not retrieve the USERID */
+		goto free_and_return;
+	}
+
+	wrap_inc_call_level();
+	if (map_rdlock() != 0) {
+		slapi_log_error(SLAPI_LOG_FATAL, state->plugin_desc->spd_id,
+				"backend_passwdmod_extop unable to acquire read lock\n");
+		wrap_dec_call_level();
+		goto free_and_return;
+	}
+	backend_locate(pb, &data, &entry_group, &entry_set);
+	if (data != NULL) {
+		/* If there is a mapping to a real entry
+		 * ndn will contains its DN
+		 */
+		if (slapi_sdn_get_ndn(data->original_entry_dn)) {
+			ndn = slapi_ch_strdup(slapi_sdn_get_ndn(data->original_entry_dn));
+		} else {
+			ndn = NULL;
+		}
+		slapi_log_error(SLAPI_LOG_PLUGIN, "backend_passwdmod_extop",
+				"reverse mapped dn = %s\n", ndn ? ndn : "<unknown>" );
+
+		/* the rest does not require to hold the map lock */
+		map_unlock();
+		wrap_dec_call_level();
+
+		if (ndn) {
+			/* replace the TARGET_SDN by the one found in the map
+			 * This is the responsibility of the extop to free it
+			 */
+			slapi_pblock_get(pb, SLAPI_TARGET_SDN, &sdn);
+			if (sdn != NULL) {
+				slapi_sdn_free(&sdn);
+			}
+			sdn = slapi_sdn_new_dn_byref(ndn);
+			slapi_pblock_set(pb, SLAPI_TARGET_SDN, (void*) sdn);
+		}
+	} else {
+		/* no mapping entry to real entry, this is fine */
+		map_unlock();
+		wrap_dec_call_level();
+	}
+
+free_and_return:
+
+	if ( ber != NULL ){
+		ber_free(ber, 1);
+		ber = NULL;
+	}
+	return 0;
+}
+static int
+backend_extop_cb(Slapi_PBlock *pb)
+{
+	struct plugin_state *state;
+	int ret = 0;
+	int i;
+	char *oid = NULL;
+	IFP fct = NULL;
+
+	slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &state);
+	if (state->ready_to_serve == 0) {
+		/* No data to serve yet */
+		return 0;
+	}
+
+	/* First check this is a supported OID (for slapi-nis) */
+	if ( slapi_pblock_get( pb, SLAPI_EXT_OP_REQ_OID, &oid ) != 0 )
+	{
+		slapi_log_error( SLAPI_LOG_FATAL, state->plugin_desc->spd_id, "Could not get OID from request\n" );
+		return 0;
+	}
+
+	for (i = 0; extop_handlers[i].oid != NULL; i++) {
+		if (strcmp( oid, extop_handlers[i].oid) == 0 ) {
+			fct = extop_handlers[i].extop_fct;
+			break;
+		}
+	}
+
+	if (fct) {
+		ret = fct(pb);
+		if (ret) {
+			slapi_log_error( SLAPI_LOG_FATAL, "backend_extop_cb",
+				 "pre-extop for %s failed %d\n", oid, ret );
+		}
+	}
+	return (ret);
+}
+
 static int
 backend_compare_cb(Slapi_PBlock *pb)
 {
@@ -2280,6 +2508,30 @@ backend_startup(Slapi_PBlock *pb, struct plugin_state *state)
 	backend_shr_startup(state, pb, SCH_CONTAINER_CONFIGURATION_FILTER);
 }
 
+void
+backend_shutdown(struct plugin_state *state)
+{
+    backend_shr_shutdown(state);
+}
+
+#ifndef SLAPI_PLUGIN_PRE_EXTOP_FN
+#define SLAPI_PLUGIN_PRE_EXTOP_FN  413
+#endif
+int
+backend_init_extop(Slapi_PBlock *pb, struct plugin_state *state)
+{
+	slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+			"hooking up extop callbacks\n");
+	/* Intercept extended operation requests */
+	if (slapi_pblock_set(pb, SLAPI_PLUGIN_PRE_EXTOP_FN,
+			     backend_extop_cb) != 0) {
+		slapi_log_error(SLAPI_LOG_PLUGIN, state->plugin_desc->spd_id,
+				"error hooking up pre extop callback\n");
+		return -1;
+	}
+	return 0;
+}
+
 int
 backend_init_preop(Slapi_PBlock *pb, struct plugin_state *state)
 {
