@@ -38,6 +38,7 @@ struct backend_set_data {
 	bool_t check_access;
 	enum sch_search_nsswitch_t check_nsswitch;
 	unsigned long nsswitch_min_id;
+        char *associated_domain;
 };
 
 struct backend_entry_data {
@@ -87,7 +88,7 @@ struct entries_to_send {
 struct backend_search_cbdata {
 	Slapi_PBlock *pb;
 	struct plugin_state *state;
-	char *target, *strfilter, **attrs;
+	char *strfilter, **attrs;
 	char *idview;
 	Slapi_Entry **overrides;
 	int scope, sizelimit, timelimit, attrsonly;
@@ -127,6 +128,22 @@ struct backend_search_filter_config {
 	void *callback_data;
 };
 
+/* OIDs of the supported extended operation */
+#define EXTOP_PASSWD_OID	"1.3.6.1.4.1.4203.1.11.1"
+
+/* ber tags for the PasswdModifyRequestValue sequence */
+#define LDAP_EXTOP_PASSMOD_TAG_USERID	0x80U
+#define LDAP_EXTOP_PASSMOD_TAG_OLDPWD	0x81U
+#define LDAP_EXTOP_PASSMOD_TAG_NEWPWD	0x82U
+
+typedef int (*IFP)();
+static int backend_passwdmod_extop(Slapi_PBlock *pb);
+typedef struct backend_extop_handlers {
+    char *oid;
+    IFP extop_fct;
+} backend_extop_handlers_t;
+
+
 /* Analyzes the filter to decide what kind of NSS search is it
  * Returns 0 on success, 1 on failure
  * struct backend_search_filter_config is populated with information about the filter
@@ -147,6 +164,7 @@ bool_t backend_retrieve_from_nsswitch(struct backend_staged_search *staged,
 				      struct backend_search_cbdata *cbdata);
 
 int backend_sch_do_pam_auth(Slapi_PBlock *pb, const char *username);
+int backend_init_extop(Slapi_PBlock *pb, struct plugin_state *state);
 
 #ifdef USE_IPA_IDVIEWS
 void idview_get_overrides(struct backend_search_cbdata *cbdata);
